from .model.compress import Compress
from .model.scan import Scan
from .model.seq_trie import SeqTrie

from .seq import Seq
from .bio_dna import BioDNA
from .dna import DNA
from .orf import ORF
from .palindromic import Palindromic

from .rna import RNA

from .prosite import Prosite
from .protein import Protein

from .trim_seq import TrimSeq



"""
"""
from typing import Iterable
from itertools import islice

class Scan:

    @staticmethod
    def forward(seq:str, start:int=None, step:int=None)->Iterable:
        if start is None or start <= 0: start = 0
        if step is None or step <= 0: step = 1
        end_pos = start
        seq = seq[start:]
        while len(seq) > 0:
            # print(seq, start)
            i = seq[:step]
            end_pos += len(i)
            seq = seq[step:]
            yield (i, end_pos)

    @staticmethod
    def backward(seq:str, step:int=None)->Iterable:
        if step is None or step <= 0: step = 1
        while len(seq) > 0:
            i = seq[-step:]
            seq = seq[:-step]
            yield i 

    @staticmethod
    def neighbor_forward(seq:str, step:int=None)->Iterable:
        '''
        return (n)th, (n+step)th at a time
        '''
        if step is None or step <= 0: step = 1
        while len(seq) > 1:
            a, b = seq[:step], seq[1:(step+1)]
            seq = seq[step:]
            yield a, b


    @staticmethod
    def biends(seq:str)->Iterable:
        while len(seq) >= 2:
            a, b = seq[0], seq[-1]
            seq = seq[1:-1]
            yield a, b
    
    @staticmethod
    def k_mers(seq:str, k:int, start:int=None) -> Iterable:
        '''
        k-mers are substrings of length k contained within a sequence
        '''
        len_seq = len(seq)
        if k <= 0:
            k = 1
        elif k > len_seq:
            k = len_seq
        if start is None or abs(start) >= len_seq:
            start = 0
        elif start < 0:
            start += len_seq
        # iteration
        for start in range(start, len_seq - k + 1):
            subseq = seq[start:start + k]
            # print(k, subseq, start, start+k-1)
            yield (subseq, start, start + k - 1)


"""
protein
"""
import re
from typing import Iterable
from Bio.Seq import Seq

from .model.scan import Scan


class Protein:
    def __init__(self, seq:str):
        self.seq = Seq(seq)
    
    def _motif(self, prosite:str, k:int):
        '''
        detect motif
        '''
        iter = Scan.k_mers(self.seq, k)
        for seq in iter:
            pass



"""
PROSITE fingerprints are described by regular expressions
https://prosite.expasy.org/prosuser.html#conv_pa
"""
import re

class Prosite:
    def __init__(self, prosite:str):
        self.prosite = prosite.upper()
        self.regexp_obj = None
    
    def motif_length(self)->int:
        '''
        determine motif length
        '''
        min_len, max_len = 0, 0
        for aa in self.prosite.split('-'):
            repeat = re.findall(r'\((\d+)\)', aa)
            if repeat:
                i = int(repeat[0])
                min_len += i
                max_len += i
            else:
                if re.findall(r'\[.*\]|\{.*\}', aa):
                    min_len += 1
                    max_len += 1
                else:
                    repeat = re.findall(r'(\w+)\((\d+,\d+)\)', aa)
                    # print(aa, repeat)
                    if repeat:
                        a,b = [int(i) for i in repeat[0][1].split(',')]
                        min_len += a * len(repeat[0][0])
                        max_len += b * len(repeat[0][0])
                    else:
                        min_len += len(aa)
                        max_len += len(aa)
        return min_len, max_len

    def compile_prosite(self):
        '''
        convert prosite to regexp and compile it into re_obj
        '''
        res = self.prosite
        res = res.replace(r'{', '[^').replace(r'}', ']')
        # repeats
        res = res.replace('(', '{').replace(')', '}')
        # N-terminal or C-terminal
        res = res.replace('<', '^').replace('>', '$')
        # any amino acids
        res = res.replace('X', r'\w')
        res = res.replace(r'-', '')
        # print(res)
        self.regexp_obj = re.compile(res)
        return self.regexp_obj

    def search_motif(self, protein:str):
        """
        search all motifs given a protein sequence
        """
        self.compile_prosite()

        if self.regexp_obj:
            protein = protein.upper()
            pool = []
            for m in re.finditer(self.regexp_obj, protein):
                # index of start, end, and matched motif
                res = (m.start(), m.end(), m.group())
                pool.append(res)
            return pool



"""
detect ORF open reading frame
ORF: protein coding region
gene sequence translated into protein
1. ORF = start_codon + sequences uninterrupted by stop codons
2. may include 5'-UTR and introns and 3'UTR.
"""
from typing import Iterable
from .dna import DNA
from .model.scan import Scan

class ORF(DNA):
    def __init__(self, seq:str=None, min_length:int=None, ignore_nested:bool=None):
        # DNA sequence
        super(ORF, self).__init__(seq)
        # minimal ORF length (nt)
        self.min_length = 30 if min_length is None else min_length
        # ignore nested ORFs
        self.ignore_nested = True if ignore_nested is None else False
        # store all ORFs
        self.orfs = []

    def is_start_codon(self, codon:str)->bool:
        '''
        '''
        start_codons = {'ATG', }
        if codon in start_codons:
            return True
        return False

    def detect_start_codon(self)->Iterable:
        '''
        search all start codon
        '''
        for codon, start, end in Scan.k_mers(self.seq, 3):
            if self.is_start_codon(codon):
                yield (codon, start, end)

    def is_termination_codon(self, codon:str)->bool:
        '''
        '''
        termination_codons = {'TAA', 'TAG', 'TGA'}
        if codon in termination_codons:
            return True
        return False

    def detect_termination_codon(self, start:int)->tuple:
        '''
        args: start=index of start codon or the nt after start codon
        '''
        for codon, end_pos in Scan.forward(self.seq, start, 3):
            # print(codon, end_pos)
            if self.is_termination_codon(codon):
                if (end_pos - start) >= self.min_length:
                    termination_codon = (codon, end_pos-3, end_pos)
                    # print(termination_codon)
                    return termination_codon
        return None
            
    def is_nested_orf(self, new_orf)->bool:
        for orf_start, orf_end in self.orfs:
            if new_orf[0]>=orf_start and new_orf[1]<=orf_end:
                return True
        return False

    def detect_orfs(self):
        '''
        suppose there is no intron.
        '''
        for start_codon in self.detect_start_codon():
            termination_codon = self.detect_termination_codon(start_codon[1])
            # print(start_codon, termination_codon)
            if termination_codon:
                orf = (start_codon[1], termination_codon[2])
                if orf not in self.orfs:
                    if (not self.ignore_nested) or (self.ignore_nested \
                        and not self.is_nested_orf(orf)):
                        self.orfs.append(orf)
        # print([(self.seq[a:b],a,b) for a,b in self.orfs])
        return self.orfs


'''
trim adapter, index, or polyA
'''
from .bio_dna import BioDNA
from .seq import Seq

class TrimSeq:
    def __init__(
        self,
        adapter:str,
        min_match:int=None,
        trim_end:str=None,
        max_err:int = None,
    ):
        self.adapter = adapter
        self.max_len = len(self.adapter)
        self.min_match = 12 if min_match is None else int(min_match)
        self.trim_end = trim_end if trim_end is not None else '3end'
        self.max_err = 0 if max_err is None else int(max_err)
        #
        seed_adapter = self.adapter[-self.min_match] if self.trim_end \
            == '5end' else self.adapter[0:self.min_match]
        self.seed_pattern = BioDNA(seed_adapter)
        self.distance = Seq(self.adapter)

    def trim_3end(self, read_seq:str)->tuple:
        match = self.seed_pattern.shift_and(read_seq, True)
        if match:
            pos = match[0]
            read_3end = read_seq[pos:pos+self.max_len]
            dist_score = self.distance.match_5end(read_3end)
            if dist_score >= len(read_3end) - self.max_err:
                return read_seq[0:pos], pos
        return read_seq, -1
        
    





"""
process DNA sequence using bioPython
"""
from Bio.Seq import Seq

class BioDNA(Seq):
    def __init__(self, seq:str=None, len:int=None):
        super(BioDNA, self).__init__(seq, len)
    
    def locate_subseq(self, sub_str:str):
        '''
        index 0-...
        '''
        return self.seq.find(sub_str)
    
    def nt_bitvector(self) -> dict:
        '''
        key-value: reversed bitset of positions
        '''
        def bitset(seq, mask, mask_len):
            if not seq:
                return mask
            letter = seq[0]
            tag = 0
            for k in list(mask):
                if k == letter:
                    mask[k] += 2 ** mask_len
                    tag = 1
                    break
            if tag == 0:
                mask[letter] = 2 ** mask_len
            return bitset(seq[1:], mask, mask_len + 1)

        #
        mask_len = self.__len__()
        mask = bitset(self, {}, 0)
        mask_str = dict([(k, format(v, f'0{mask_len}b')) for k,v in mask.items()])
        return mask, mask_str

    def format_code(self, code:int):
        return format(code, f"0{self.len}b")
  
    def shift_and(self, text:str, first_occurrence:bool=False):
        '''
        exact match pattern: Shift-and
        '''
        pattern, _ = self.nt_bitvector()
        pattern_len = self.__len__()

        p, res = 0, []
        target = 2 ** (pattern_len - 1)
        for i, v in enumerate(text):
            v_code = pattern.get(v, 0)
            p = (p << 1 | 1) & v_code
            # print(self.format_code(v_code), self.format_code(p))
            if p >= target:
                res.append(i - pattern_len + 1)
                if first_occurrence:
                    return res
        return res


"""
process sequence
"""
from typing import Iterable, Callable

class Seq:
    def __init__(self, seq:str=None):
        self.seq = seq.upper().replace('\n', '') if \
            seq and isinstance(seq, str) else ''
    
    def length(self)->int:
        return len(self.seq)

    def reverse(self)->str:
        return self.seq[::-1]
    
    def count_sub_seq(self, sub_seq:str)->list:
        if len(sub_seq)==0:
            return 0
        return self.seq.count(sub_seq)

    def count_occurrence(self)->dict:
        res = {}
        for i in self.seq:
            if i in res:
                res[i] += 1
            else:
                res[i] = 1
        return res
    
    def search_sub_seq(self, sub_seq:str)->list:
        '''
        search all subsequences 
        '''
        if len(sub_seq)==0:
            return []
        #
        count, start = [], 0
        while start != -1:
            pos = self.seq.find(sub_seq, start)
            if pos >= 0:
                end = pos + len(sub_seq)
                count.append((pos, end))
                start += end
            else:
                start = -1
        return count
        
    def calculate_hamming_distance(self, seq2:str, func:Callable=None)->int:
        '''
        number of positions of that two codewords of the same length differe
        Note: self.seq and seq2 may be same lengths
        '''
        if func is None:
            def func(a, b):
                return a != b
        #
        dist = 0
        for a,b in zip(self.seq, seq2):
            if func(a, b):
                dist += 1
        dist += abs(len(self.seq)-len(seq2))
        return dist
        
    def calculate_similarity(self, seq2:str)->float:
        '''
        Note: self.seq and seq2 may be same lengths
        '''
        dist = self.calculate_hamming_distance(seq2)
        seq_len = len(self.seq) if len(self.seq) >= len(seq2) else len(seq2)
        return (seq_len - dist)/seq_len

    def match_5end(self, seq2:str) -> int:
        '''
        match from 5'-end
        return distance score
        '''
        dist = 0
        end = len(seq2) if len(seq2) <= len(self.seq) else len(self.seq)
        for i in range(0, end):
            if self.seq[i] == seq2[i]:
                dist += 1
        return dist




"""
process DNA sequence using naive methods
"""
from typing import Iterable
from .dna import DNA
from .model.scan import Scan

class Palindromic(DNA):

    def __init__(self, seq:str=None):
        super(Palindromic, self).__init__(seq)

    def detect_longest_palindromic(self)->list:
        if self.length() > 2:
            for palindromic_len in range(self.length(), 2, -1):
                palindromic = self.detect_palindromic(palindromic_len)
                if palindromic:
                    return palindromic
        return []

    def detect_palindromic(self, palindromic_len:int)->list:
        '''
        given sequence length, 
        '''
        pool = []
        iter = Scan.k_mers(self.seq, palindromic_len)
        for sub_seq in iter:
            if Palindromic.is_palindromic(sub_seq[0]):
                pool.append(sub_seq)
        return pool

    @staticmethod
    def is_palindromic(seq:str, iter:str=None)->bool:
        if iter is None: iter = 0
        if len(seq) >= 2:
            first, last = seq[0], seq[-1]
            if first == last:
                return Palindromic.is_palindromic(seq[1:-1], iter+1)
            return False
        else:
            if iter == 0:
                return False
        return True




'''
Test ENV
'''
from tests.helper import *


@ddt
class TestENV(TestCase):

    def setUp(self):
        pass

    @mock.patch.dict(os.environ, env)
    def test_env(self):
        print(os.getenv('DIR_DOWNLOAD'))


"""
process DNA sequence using naive methods
"""
from typing import Iterable
from .seq import Seq
from .model.scan import Scan

class DNA(Seq):
    nt_pair = {'A':'T', 'T':'A', 'G':'C', 'C':'G',}

    def __init__(self, seq:str=None):
        super(DNA, self).__init__(seq)
    
    def complement(self)->str:
        return DNA.replace(self.seq)

    def reverse_complement(self)->str:
        return DNA.replace(self.seq[::-1])

    @staticmethod
    def replace(seq:str):
        if len(seq)==0:
            return ''
        nt = DNA.nt_pair.get(seq[0], seq[0])
        return nt + DNA.replace(seq[1:])

    def calculate_gc(self):
        '''
        GC percentage
        '''
        if self.length() == 0:
            return 0
        g = self.count_sub_seq('G')
        c = self.count_sub_seq('C')
        return (g + c)/self.length()
    
    def detect_similarity(self, seq2:str)->tuple:
        '''
        Note: length of seq2<= self.seq
        '''
        len_seq = len(seq2)
        min_dist, pool = len_seq, []
        S = Seq(seq2)
        for seq1, start, end in Scan.k_mers(self.seq, len_seq):
            dist = S.calculate_hamming_distance(seq1)
            if dist < min_dist:
                min_dist = dist
                pool = [(start, end)]
            elif dist == min_dist:
                pool.append((start, end))
        max_same = len_seq - min_dist
        return max_same, pool

    def detect_overlap(self, seq2:str)->tuple:
        '''
        3-end of seq1 is overlapped with 5-end of seq2
        '''
        olen = 0
        while self.seq[-(olen+1):] == seq2[:(olen+1)]:
            olen +=1
        return olen

    def detect_target_repeat(self, repeat:str=None)->list:
        if repeat is None: repeat = 'N'

        pool = []
        start = self.seq.index(repeat)
        end = start + len(repeat)
        while 0 <= start < self.length():
            next_seq = self.seq[start+1:]
            pos = next_seq.index(repeat)
            if pos == 0:
                end += len(repeat)
            elif pos > 0:
                pool.append((start, end))
                start = pos
                end = start + len(repeat)
            else:
                break
        return pool
    
    def detect_longest_target_repeat(self, repeat:str=None)->tuple:
        pool = self.detect_target_repeat(repeat)
        repeat_start, repeat_end = 0, 0
        for start, end in pool:
            if (end - start) > (repeat_end - repeat_start):
                repeat_start, repeat_end = start, end
        return repeat_start, repeat_end

        


from setuptools import setup, find_packages
import codecs
import os

here = os.path.abspath(os.path.dirname(__file__))

with codecs.open(os.path.join(here, "README.md"), encoding="utf-8") as fh:
    long_description = "\\n" + fh.read()

setup(
    name="eseq",
    version='0.0.4',
    author="Tiezheng Yuan",
    author_email="tiezhengyuan@hotmail.com",
    description="Process genomic sequences.",
    url = "https://github.com/Tiezhengyuan/bio_sequence.git",
    long_description_content_type="text/markdown",
    long_description=long_description,
    package_dir={'': 'src'},
    install_requires=['Bio',],
    keywords=['pypi', 'cicd', 'python'],
    classifiers=[
        "Development Status :: 1 - Planning",
        "Intended Audience :: Developers",
        "Programming Language :: Python :: 3",
        "Operating System :: Unix",
        "Operating System :: MacOS :: MacOS X",
        "Operating System :: Microsoft :: Windows"
    ]
)



"""
compress sequence
"""
import re
from typing import Iterable

class Compress:
    def __init__(self):
        pass
    
    @staticmethod
    def encode_repeat(seq:Iterable)->Iterable:
        '''
        count consecutive repeating characters as digits
        '''
        if len(list(seq)) == 0:
            return []
        encoded = []
        curr, count = seq[0], 1
        for i in seq[1:]:
            if i != curr:
                encoded.append((curr, count))
                curr, count = i, 1
            else:
                count += 1
        else:
            encoded.append((curr, count))
        return encoded

    @staticmethod
    def to_encoded_seq(seq:Iterable)->str:
        encoded = Compress.encode_repeat(seq)
        encoded_seq = []
        for nt, count in encoded:
            encoded_seq.append(nt)
            if count > 1:
                encoded_seq.append(str(count))
        return ''.join(encoded_seq)
            

    @staticmethod
    def decode_repeat(encoded_seq:Iterable):
        encoded = filter(None, re.split(r'([A-Za-z])', encoded_seq))
        seq = []
        for i in encoded:
            try:
                seq[-1] = seq[-1] * int(i)
            except:
                seq.append(i)
        return ''.join(seq)

                    



"""
split sequence into nodes
basic unit of a trie
"""
class Node:
    def __init__(self, val:str=None):
        self.val = val if val is not None else ''
        self.is_leave = False
        # distance between this node to root node
        self.depth = 0
        #root node has not father
        # other nodes have only one father
        self.father = None
        # leave nodes have no children
        # other nodes have one or many children
        self.children = {}
        # the two attr used for locating seq
        # index of this node in the entire sequence
        self.val_pos = ()

    def child_node(self, child_val:str):
        children = getattr(self, 'children')
        if child_val in children:
            return children[child_val]
        return None

    def father_node(self):
        return getattr(self, 'father')
    
    def is_leave_node(self)->bool:
        return self.is_leave


"""
store long sequence into a trie
slice long sequence into many seqs
one seq is stored into the trie
"""
from typing import Iterable
from .scan import Scan

class SeqTrie:
    def __init__(self, root_val:str=None):
        # mostly, root node has no value
        # sometimes all sequences share identifical prefix
        self.root = Node(root_val)
        self.seq_nodes = ()

    def insert(self, seq:str, end_pos:int=None)->Node:
        '''
        args: val is sequence namely A/T/G/C
        '''
        # print('###', seq)
        curr_node = self.root
        for val in seq:
            if val not in curr_node.children:
                # print(val)
                this_node = Node(val)
                this_node.depth = curr_node.depth + 1
                this_node.father = curr_node
                curr_node.children[val] = this_node
                curr_node = this_node
            else:
                curr_node = curr_node.children[val]
        else:
            curr_node.is_leave = True
            if end_pos:
                curr_node.val_pos += (end_pos,)
        self.seq_nodes += (curr_node,)
        return curr_node

    def dfs_search(self, curr_node:Node, prefix:str)->Iterable:
        '''
        retrieve entrie sequence from trie
        '''
        # print(prefix, curr_node.val)
        prefix += curr_node.val
        if curr_node.is_leave:
            yield prefix, curr_node.val_pos
        else:
            for child_node in curr_node.children.values():
                yield from self.dfs_search(child_node, prefix)

    def scan(self)->Iterable:
        '''
        retrieve sequence slices from trie
        Note: order of slices may differ from origin sequence
        '''
        curr_node = self.root
        for seq, _ in self.dfs_search(curr_node, ''):
            yield seq


    def get(self, leave_node:Node)->str:
        '''
        get seq slice based on reference of leave node
        '''
        seq = ''
        curr_node = leave_node
        while hasattr(curr_node, 'father'):
            seq = curr_node.val + seq
            curr_node = curr_node.father
        return seq
    
    def retrieve(self)->Iterable:
        '''
        retrieve seq
        '''
        for leave_node in self.seq_nodes:
            yield self.get(leave_node)

    def insert_sequence(self, sequence:str, layers:int=None):
        '''
        split long sequence into slices which are inserted into Trie
        '''
        if layers is None or layers < 1: layers = 10
        for seq, end_pos in Scan.forward(sequence, 0, layers):
            self.insert(seq, end_pos)

    def get_sequence(self)->str:
        '''
        retrieve entrie sequence from trie
        Note: not recommended if sequence is long
        '''
        sequence = ()
        for leave_node in self.seq_nodes:
            sequence += (self.get(leave_node),)
        return ''.join(sequence)



